#ifndef _PW_PWSYSTEM_H
#define _PW_PWSYSTEM_H

/*
	ProWesS, PROGS Windowing System

	Definition of the internal system

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	started 7 January 1995
	20.03.1996 reverted to using pixel coordinates, should be faster
*/

#include <mem.h>
#include <win.h>
#include <job.h>

#ifndef _PW_PWEVENT_H
#include "PWEvent.h"
#endif

/* Each type can have a general state, to keep information about it */
/* Normally the base type is followed by extra type specific information */
/* The PWTypeState information is kept in a list with sentinel (type=NULL) */

#define PWBaseTypeState 									\
	PWType type;					/* type identifier */	\
	struct _PWDummyTypeState *next	/* next item */

typedef struct _PWDummyTypeState {
	PWBaseTypeState;				/* common start for all typestates */
} PWDummyTypeState;

/* System data structure */

#define PWPublicSystem														  \
	unsigned long status;			/* system status (values SSTATUS_xxx) */  \
																			  \
	Gstate gstate;					/* PROforma gstate is used (or NULL) */   \
	Window window;					/* Window identifier */ 				  \
																			  \
	PWObject catchkeys; 			/* object which has to catch keypresses */\
	char catch; 					/* key for which catching is needed */	  \
	char activated; 				/* TRUE when system activated */		  \
																			  \
	short ptr_xpos, ptr_ypos;		/* current pointer position */			  \
	pix win_xpos, win_ypos; 		/* position of window */				  \
	pix xorg, yorg; 				/* origin of area in window */			  \
																			  \
	Heap heap;						/* heap with/for this system */ 		  \
																			  \
	int Xdragadjust, Ydragadjust;	/* adjustment of pointer while dragging */\
	int Xspaceleft, Yspaceleft		/* space left for auto increment of obj */

#ifndef _PW_PRIVATE
struct _PWSystem {
	PWPublicSystem;
};
#endif _PW_PRIVATE

#define SSTATUS_REDRAW		(unsigned long)0x00000001  /* set when part of window needs redraw */
#define SSTATUS_WINCHANGE	(unsigned long)0x00000004  /* set when window build has changed */
#define SSTATUS_PAGESHOW	(unsigned long)0x00000040  /* set when PFPageShow should be done */

#endif _PW_PWSYSTEM_H
