; some subroutines for the Sbasic interface to Prowess
; they get the parameters & push them onto the A4 stack
; v. 1.00 1996 May 06 13:52:51
; copyright (c) W. Lenerz 1996

	section text

; these routines get parameters from the Sbasic interpreter: all parameters are
; either strings or treated as long integers (!)

	xdef	get_long		; gets one parameter
	xdef	get_param
	xdef	get_all		; get all parameters & push onto stack
	xdef	get_all2		; same but does not put NULL as first param
	xdef	get_para

; entry				exit

; D0 ?				error return
; D1 - D5 ?			preserved
; D6 ?				parameter got (!) (if LONG) <<<<<<<
; A0 ?				preserved
; A1.L	basic stack		same
; A2 ?				<corrupted>
; A3	pointer to start of param	pointer to start of next parameter
; A4	top of stack		new top of stack after param has been put there
; A5	pointer to end of param	pointer to end of next parameter
; A6	basic data pointer		same



; if there is an error, a direct return is made from the subroutine
; to my caller

; -----------------------------------------------------------------
; get_long: get a parameter as LONG integer and push it onto the stack

get_long	BSR.S	get_para		; (if error direct return to caller)
	MOVE.L	(A6,A1.L),D6
	MOVE.L	D6,-(A4)		; put parameter directly onto stack
	TST.L	D0		;
	RTS

;----------------------------------------------------------------------------
; this actually gets ONE parameter, a long integer
; if error, it returns directly to the routine having called this one
; on entry A3 and A5 must point to ONE parameter, as usual
; the parameter is left at (a6,a1.l)
; the A3 and A5 parameter pointers are adjusted to point to the parameter
; BEFORE this one

grregs	REG	D1-D4/D6/A0
get_para	MOVEM.L	 grregs,-(A7)	; keep
	MOVE.W	$118,A2
	JSR	(A2)		; get long integer, this converts all other types...
	BNE.S	err_bpm		; oops->...
	MOVEQ	#-15,D0		; preset error
	SUBQ.B	#1,D3		; there must be exactly 1 parameter
	BNE.S	err_bpm		; oops->...
	MOVEM.L	(A7)+,grregs
	ADD.L	#4,$58(A6)	; reset maths stack
	MOVE.L	A3,A5		; point to next parameter
	SUBQ.L	#8,A3		; same here
	MOVEQ	#0,D0		; no error
	RTS
err_bpm	MOVEM.L	(A7)+,grregs	; error return
	ADDQ.L	#4,A7
	TST.L	D0		; skip return address !
	RTS



; this just calls get_para (necessary since get_para skips return address in
; case of error)
get_param BSR.S	get_para
	RTS

*------------------------------------------------------------------------
; this gets all parameters and puts them onto the stack

err_out	MOVEQ	#-15,D0
out	RTS
get_all	MOVE.L	#0,-(A4)	; last param on stack is always NULL
get_all2	MOVE.L	 A5,D5		 ; point to end of last param...
	SUB.L	A3,D5		; ...minus start of first = nbr of params
	BEQ.S	err_out		; there must be at least one!
	LSR.W	#3,D5		; number of prarams now
	SUBQ.W	#1,D5		; prepare for DBF
	LEA	-8(A5),A3 ; point to last parameter
gpar_lp1	MOVE.W	 (A6,A3.L),D0	 ; type word into D0
	ANDI.B	#$F,D0		; get rid of interference
	CMP.B	#1,D0		; is it a string?
	BEQ.S	is_str		; yes ->....

; here parameter is not a string, so it is a long word
no_str	BSR	get_long	; get the long word
	BNE.S	err_out		; oops->...
	DBF	D5,gpar_lp1	; and do the loop again
	BRA.S	continue	; after loop is finished


; here the parameter to get is a string  - see rules for passing strings below
is_str	MOVE.W	2(A6,A3.L),D2
	MOVE.L	4(A6,A3.L),A1	; point to value now (this is Sbasic, so no indexing rel to A6 here!)
	CMP.W	#-1,D2		; is it an unset string (function etc)
	BEQ.S	unset		; yes->..., so stock string temporarily
	CMP.W	#$201,D0		; is it array?
	BEQ.S	on_stack		; no->...
	MOVE.L	(A1),A1		; true value pointer if array
on_stack	ADDQ.L	#2,A1		; point after length word
	MOVE.L	A1,-(A4)		; point to this on stack
	BRA.S	end_str

; now put the string onto the BOTTOM of our stack area, and the address of this
; onto the top of the stack... There is a risk of the top and the bottom meeting,
; but this would take an extraordinarily long parameter list! (4 K long)

unset	MOVE.W	(A1)+,D0	; length of string
	MOVE.W	D0,(A0)+	; this isn't always necessary, but may be for config
	MOVE.L	A0,-(A4)	; this is where the string lies
	BRA.S	copy_str
str_lp1	MOVE.B	(A1)+,(A0)+
copy_str DBF	D0,str_lp1
	MOVE.B	#0,(A0)+	; end with byte 0 for 'c' string
	MOVE.L	A0,D0
	ADDQ.L	#1,D0
	BCLR	#0,D0
	MOVE.L	D0,A0		; make sure A0 is even
end_str	MOVE.L	A3,A5		; point to previous parameter
	SUBQ.L	#8,A3
do_loop1 DBF	D5,gpar_lp1	; and get it now
continue MOVEQ	#0,D0
	RTS

	end
