/*
	ProWesS, PROGS Windowing System

	PWDefine, define the builtin parameters

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	started 14 January 1995
	09.04.1996 - added default defines for foreground colour, font and fontsize
			   - completely restructed, introduced aid routine for parsing
			   - SYSTEM-FONT-CALCULATED added
*/

#include <str.h>

#include "ProWesS.h"
#include "PWHandler.h"

#define p100 short2pt(100)

short	  system_xshadow=6;
short	  system_yshadow=6;
short	  system_borderwidth=1;
short	  system_bordercolour=0;
pt		  system_scroll_distance=short2pt(20);
short	  system_scaleborderwidth=4;
short	  system_scalebordercolour=2;
char	  system_previewmove=TRUE;
char	  system_previewscale=TRUE;
Gstate	  system_gstate=NULL;
short	  system_previewtimeout=10;
TimeOut   system_dragtimeout=11;
ColourRGB system_paper={p100,p100,p100};
ColourRGB system_ink={0,0,0};
ColourRGB system_middle={0,p100,0};
pt		  system_fontsize=short2pt(22);
char	  system_font[PF_MAXFONTNAME]="??";

void Even(void *vdata)
{
	short *data=(short *)vdata;
	*data=(*data+1)&-2;
}

static Error LoadResident(Defines *p, char *value)
{
	return PFFontLoad(system_gstate,value);
}

static Error CalculateFont(Defines *p, char *value)
{
	double size; int len;
	STRFormatGet(value,NULL,"%f %n",&size,&len);
	PFFontLoad(system_gstate,value+len);
	PFFontScale(system_gstate,double2pt(size));
	return PFFontCalculate(system_gstate);
}

Defines defines[]={
{ "SHADOW-RIGHT",       PW_DEFINE_SHORT,    Even,   &system_xshadow,            NULL},
{ "SHADOW-BOTTOM",      PW_DEFINE_SHORT,    NULL,   &system_yshadow,            NULL},
{ "BORDER-WIDTH",       PW_DEFINE_SHORT,    NULL,   &system_borderwidth,        NULL},
{ "BORDER-COLOUR",      PW_DEFINE_SHORT,    NULL,   &system_bordercolour,       NULL},
{ "SCALEBORDER-WIDTH",  PW_DEFINE_SHORT,    Even,   &system_scaleborderwidth,   NULL},
{ "SCALEBORDER-COLOUR", PW_DEFINE_SHORT,    NULL,   &system_scalebordercolour,  NULL},
{ "PREVIEW-MOVE",       PW_DEFINE_BOOLEAN,  NULL,   &system_previewmove,        NULL},
{ "PREVIEW-SCALE",      PW_DEFINE_BOOLEAN,  NULL,   &system_previewscale,       NULL},
{ "PREVIEW-TIMEOUT",    PW_DEFINE_SHORT,    NULL,   &system_previewtimeout,     NULL},
{ "LOAD-RESIDENT-FONT", LoadResident,       NULL,   NULL,                       NULL},
{ "FONT-CALCULATED",    CalculateFont,      NULL,   NULL,                       NULL},
{ "SCROLL-DISTANCE",    PW_DEFINE_PT,       NULL,   &system_scroll_distance,    NULL},
{ "DRAGTEST-TIMEOUT",   PW_DEFINE_SHORT,    NULL,   &system_dragtimeout,        NULL},
{ "BACKGROUND-COLOUR",  PW_DEFINE_COLOUR,   NULL,   &system_paper,              NULL},
{ "FOREGROUND-COLOUR",  PW_DEFINE_COLOUR,   NULL,   &system_ink,                NULL},
{ "MIDDLEGROUND-COLOUR",PW_DEFINE_COLOUR,   NULL,   &system_middle,             NULL},
{ "FONTSIZE",           PW_DEFINE_PT,       NULL,   &system_fontsize,           NULL},
{ "FONT",               PW_DEFINE_FONT,     NULL,   &system_font,               NULL},
{ 0 }
};


Error PWDefine(char *name, char *value)
{
	return PWHandleDefine("SYSTEM-",defines,name,value);
}

Error PWHandleDefine(char *prefix, Defines *possible, char *name, char *value)
{
	Error err=ERR_OK;

	/* start by testing for precense of the prefix */
	while (*prefix) if (*prefix++!=*name++) return ERR_IPAR;

	/* search for a matching name */
	while (possible->name && !STRSameCD(possible->name,name)) possible++;

	/* handle when found */
	if (!possible->name) return ERR_IPAR;
	if ((int)possible->Handle>50)
		err=possible->Handle(possible,value);
	else
	switch((int)possible->Handle)
	{
	case PW_DEFINE_COLOUR:
		{
			double red,green,blue;
			STRFormatGet(value,NULL,"%f %f %f",&red,&green,&blue);
			((ColourRGB *)possible->data)->red	=double2pt(red);
			((ColourRGB *)possible->data)->green=double2pt(green);
			((ColourRGB *)possible->data)->blue =double2pt(blue);
		}
		break;
	case PW_DEFINE_BOOLEAN:
		*(char *)possible->data=FALSE;
		if (*value=='t' || *value=='T') *(char *)possible->data=TRUE;
		break;
	case PW_DEFINE_CHARACTER:
		*(char *)possible->data=*value;
		break;
	case PW_DEFINE_SHORT:
		STRFormatGet(value,NULL,"%hd",(short *)possible->data);
		break;
	case PW_DEFINE_INT:
		STRFormatGet(value,NULL,"%d",(int *)possible->data);
		break;
	case PW_DEFINE_PT:
		{
			double db;
			STRFormatGet(value,NULL,"%f",&db);
			*(pt *)possible->data=double2pt(db);
		}
		break;
	case PW_DEFINE_FONT:
		STRCopySafe(value,(char *)possible->data,PF_MAXFONTNAME);
		break;
	default:
		return ERR_IPAR;
	}
	if (possible->Post) possible->Post(possible->data);
	if (possible->set) *possible->set=TRUE;
	return err;
}
