/*
	ProWesS, PROGS Windowing System

	PWDetermineSize

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	started 14 January 1995
	26.10.1995 PWDetermineSize now works in a loop, the scaleborder will be
			   drawn for each pass in the loop - as long as it works !
	15.12.1995 fixed problem with determining position for window
	02.01.1996 scaleborder should now be draw correctly always, also right
			   and bottom scaleborder (was sometimes pixel too far outwards)
	20.03.1996 some internal structures are now in pixels (not pt)
	23.03.1996 slight improvement by integrating DetermineSizeRegion and
			   finding maximum screen size in one loop + removed DefaultDetSize
	24.03.1996 moved some code to PWWindowFit_c
	24.07.1996 in button frame, window xsize now properly forced to be (?*4)
	14.12.1996 in button frame, window ysize now properly forced to be (?*2)
	15.04.1997 - save pointer pos for restore when closing window
			   - restructured to allow dynamic sizing of windows and setting
				 the pointer position
	21.04.1997 setting pointer position in system now also works when window
			   is larger than screen
*/

#include <win.h>
#include <thing.h>

#include "private.h"
#include "PWHandler.h"

/* forward declarations */
static Error PWDetermineSizeRegion(PWSystem *system, PWRegion *region);
static Error PWPrepareWindow(PWSystem *system);

/* external definitions */
extern short  system_xshadow;
extern short  system_yshadow;
extern short  system_borderwidth;
extern short  system_bordercolour;
extern pt	  Xpixel2point, Ypixel2point;
extern short  system_scaleborderwidth;
extern short  system_scalebordercolour;
extern Gstate system_gstate;

#define catch(X) if (err=(X)) return err; else

Error PWDetermineSize(PWSystem *system)
{
	PWRegion *region;
	Error err;
	pt xmax=0, ymax=0;
	WindowInfo wininfo;

	while (system->status&SSTATUS_WINCHANGE)
	{
		system->status&=~SSTATUS_WINCHANGE;

		system->status|=SSTATUS_ALLDRAW;	/* new sizes -> so needs redraw */

		/* determine the size of all the regions */
		/* AND determine the (minimum) size of the main window */
		/* we have to make sure that all the children fit inside */
		for (region=system->regions; region; region=region->next)
		{
			catch( PWDetermineSizeRegion(system, region) );
			xmax=max(xmax,region->tot_xpos+region->tot_xsiz);
			ymax=max(ymax,region->tot_ypos+region->tot_ysiz);
		}
		system->min_xsiz=pt2rshort(xmax);
		system->min_ysiz=pt2rshort(ymax);

		/* and the maximum size of course */
		if (system->window==WINDOW_NONE)
			catch( WINOpen(2,1,0,0,&system->window) );
		catch( WINLimits(system->window,&wininfo) );
		system->max_xsiz=wininfo.xsiz-system_xshadow;
		system->max_ysiz=wininfo.ysiz-system_yshadow;

		catch( PWPrepareWindow(system) );

		/* do windowfitting */
		PWWindowFit(system,short2pt(max(0,system->curr_xsiz-system->min_xsiz)),
						   short2pt(max(0,system->curr_ysiz-system->min_ysiz)));

		/* allow autosize of one object */
		if (system->autosize)
		{
			system->Xspaceleft=system->max_xsiz-
					max(system->curr_xsiz,system->min_xsiz)-2*system->borderx;
			system->Yspaceleft=system->max_ysiz-
					max(system->curr_ysiz,system->min_ysiz)-2*system->bordery;
			if (system->autosize->type->PWHandle)
			{
				system->autosize->type->PWHandle(system->autosize,PW_EVENT_AUTOSIZE);
			}
		}
	}
	return PWSetWindow(system);
}

static Error PWDetermineSizeRegion(PWSystem *system, PWRegion *region)
{
	PWRegion *child;
	pt xmax=0, ymax=0;
	Error err;

	for (child=region->children ; child ; child=child->next)
	{
		catch( PWDetermineSizeRegion(system,child) );
		xmax=max(xmax,child->tot_xpos+child->tot_xsiz);
		ymax=max(ymax,child->tot_ypos+child->tot_ysiz);
	}
	/* set size and use margins (if initialised in Create ?) */
	region->tot_xsiz=xmax+short2pt(region->hit_off_lft+region->hit_off_rgt);
	region->tot_ysiz=ymax+short2pt(region->hit_off_top+region->hit_off_bot);

	if (region->owner && region->owner->type->PWDetermineSize)
		catch( region->owner->type->PWDetermineSize(region->owner) );

	/* set position of total area of region */
	region->tot_xpos=region->tot_ypos=0;
	if (region->prev)
	{
		if (region->level&1)
			region->tot_xpos=region->prev->tot_xpos+region->prev->tot_xsiz;
		else
			region->tot_ypos=region->prev->tot_ypos+region->prev->tot_ysiz;
	}
	return ERR_OK;
}

void PWDeterminePosition(PWSystem *system)
{
	WindowInfo wi;
	int pxsiz=system->curr_xsiz;
	int pysiz=system->curr_ysiz;
	int wxpos=system->win_xpos;
	int wypos=system->win_ypos;

	WINLimits(system->window,&wi);
	WINPointerPositionGet(system->window,&system->oldptrx,&system->oldptry);

	if (wxpos==-1) wxpos=system->oldptrx-system->win_ptrx;
	if (wypos==-1) wypos=system->oldptry-system->win_ptry;

	wxpos=max(wxpos,wi.xpos);
	wypos=max(wypos,wi.ypos);
	wxpos=min(wxpos,((wi.xsiz+wi.xpos-system->borderx*2-system_xshadow)-pxsiz) );
	wypos=min(wypos,((wi.ysiz+wi.ypos-system->bordery*2-system_yshadow)-pysiz) );

	if (system->win_xpos==-1 || system->win_ypos==-1 || system->status&SSTATUS_BUTFRAME)
	{
		system->xorg=system->yorg=0;
		if (system->win_ptrx>pxsiz)
			system->xorg=min( system->win_ptrx-pxsiz/2,
							  system->min_xsiz-pxsiz );
		if (system->win_ptry>pysiz)
			system->yorg=min( system->win_ptry-pysiz/2,
							  system->min_ysiz-pysiz );
		WINPointerPositionSet(system->window,
							  wxpos+system->win_ptrx+system->borderx-system->xorg,
							  wypos+system->win_ptry+system->bordery-system->yorg);
		system->xorg-=system->borderx;
		system->yorg-=system->bordery;
	}
	system->win_xpos=wxpos&-2;
	system->win_ypos=wypos;
}

Error PWSetSprites(PWSystem *system)
{
	SubWindow *sw;
	int xsiz=system->curr_xsiz;
	int ysiz=system->curr_ysiz;
	int borders=system_borderwidth+system_scaleborderwidth;
	int borderwidth=borders+system_borderwidth;
	int dborderwidth=borderwidth*2;
	int extraborder=borderwidth+system_borderwidth;
	int i;

	/* set subwindows */

	sw=&system->subwindows.main;
	sw->xsiz=xsiz+dborderwidth;
	sw->ysiz=ysiz+dborderwidth;
	sw->xpos=system->win_xpos;
	sw->ypos=system->win_ypos;

	if (system->status&SSTATUS_NOXBORDER) xsiz-=dborderwidth;
	if (system->status&SSTATUS_NOYBORDER) ysiz-=dborderwidth;

	/* 0 top-left corner - horizontal bit */
	/* 1 top-left corner - vertical bit */
	/* 2 top-right corner - horizontal bit */
	/* 3 top-right corner - vertical bit */
	/* 4 bottom-left corner - horizontal bit */
	/* 5 bottom-left corner - vertical bit */
	/* 6 bottom-right corner - horizontal bit */
	/* 7 bottom-right corner - vertical bit */
	sw=system->subwin;
	sw[0].xsiz=sw[2].xsiz=sw[4].xsiz=sw[6].xsiz=
					system_scaleborderwidth+(xsiz+system_borderwidth)/2;
	sw[0].ysiz=sw[2].ysiz=sw[4].ysiz=sw[6].ysiz=
	sw[1].xsiz=sw[3].xsiz=sw[5].xsiz=sw[7].xsiz=system_scaleborderwidth;
	sw[1].ysiz=sw[3].ysiz=sw[5].ysiz=sw[7].ysiz=
					(ysiz+system_borderwidth)/2;
	sw[0].xpos=sw[1].xpos=sw[4].xpos=sw[5].xpos=
	sw[0].ypos=sw[2].ypos=system_borderwidth;
	sw[1].ypos=sw[3].ypos=borders;
	sw[2].xpos=sw[6].xpos=xsiz+borderwidth+borders-sw[2].xsiz;
	sw[3].xpos=sw[7].xpos=xsiz+extraborder;
	sw[4].ypos=sw[6].ypos=ysiz+extraborder;
	sw[5].ypos=sw[7].ypos=ysiz+borderwidth+system_borderwidth-sw[5].ysiz;

	if (system->status&SSTATUS_NOXBORDER)
	{
		for (i=1;i<8;i+=2)
			sw[i].xpos=sw[i].ypos=sw[i].xsiz=sw[i].ysiz=0;
	}
	if (system->status&SSTATUS_NOYBORDER)
	{
		for (i=0;i<8;i+=2)
			sw[i].xpos=sw[i].ypos=sw[i].xsiz=sw[i].ysiz=0;
	}

	return WINSubWindows(system->window,&system->subwindows);
}

Error PWDrawBorders(PWSystem *system)
{
	Window win=system->window;
	int borderx=system->borderx;
	int bordery=system->bordery;
	SubWindow *sw=system->subwin;
	int xsiz=system->curr_xsiz;
	int ysiz=system->curr_ysiz;
	int i;

	/* normal border */
	WINBlock(win,system_bordercolour, xsiz+borderx*2, bordery, 0, 0);
	WINBlock(win,system_bordercolour, borderx, ysiz+bordery*2, 0, 0);
	WINBlock(win,system_bordercolour, xsiz+borderx*2, bordery, 0, ysiz+bordery);
	WINBlock(win,system_bordercolour, borderx, ysiz+bordery*2, xsiz+borderx,0);
	/* scale piece inside border */
	
	PWSetSprites(system);
	for (i=0;i<8;i++,sw++)
	{
		WINBlock(win,system_scalebordercolour,
				 sw->xsiz, sw->ysiz, sw->xpos, sw->ypos);
	}
	return ERR_OK;
}

/* determine actual size of window */
static Error PWPrepareWindow(PWSystem *system)
{
	SubWindow *sw=system->subwin;
	int xsiz=system->curr_xsiz;
	int ysiz=system->curr_ysiz;
	int i;
	Window win=system->window;

	system->borderx=system->bordery=system_scaleborderwidth+system_borderwidth*2;

	if (system->status&SSTATUS_NOXSCALE) xsiz=system->min_xsiz;
	if (system->status&SSTATUS_NOYSCALE) ysiz=system->min_ysiz;

	/* clear statusses before we possibly set them */
	system->status&=~(SSTATUS_NOXBORDER | SSTATUS_XSCROLLABLE |
					  SSTATUS_NOYBORDER | SSTATUS_YSCROLLABLE);
	/* determine size of current window */
	if (system->min_xsiz > system->max_xsiz-system->borderx*2)
	{
		system->status|=SSTATUS_NOXBORDER|SSTATUS_XSCROLLABLE;
		xsiz=system->max_xsiz;
	}
	if (system->min_ysiz > system->max_ysiz-system->bordery*2)
	{
		system->status|=SSTATUS_NOYBORDER|SSTATUS_YSCROLLABLE;
		ysiz=system->max_ysiz;
	}

	if (system->status&SSTATUS_NOXBORDER) system->borderx=0;
	if (system->status&SSTATUS_NOYBORDER) system->bordery=0;

	if (system->min_xsiz				  >xsiz) xsiz=system->min_xsiz;
	if (system->min_ysiz				  >ysiz) ysiz=system->min_ysiz;
	if (system->max_xsiz-system->borderx*2<xsiz) xsiz=system->max_xsiz-system->borderx*2;
	if (system->max_ysiz-system->bordery*2<ysiz) ysiz=system->max_ysiz-system->bordery*2;
	xsiz=(xsiz+1)&-2;

	if (system->status&SSTATUS_BUTFRAME)
	{
		Error err;
		int butret,siz;
		/* make button frame size proper (size should be  ?x4,?x2) */
		if ((xsiz+system_borderwidth*2)&2) xsiz+=2;
		if ((ysiz+system_borderwidth*2)&1) ysiz+=1;
		/* try to allocate room in the button frame */
		err=THINGUseFull("Button Frame",
						 ((xsiz+system_borderwidth*2)<<16) +
						   ysiz+system_borderwidth*2,
						 MYSELF,0,0,&butret,NULL,NULL);
		if (err)
			system->status&=~SSTATUS_BUTFRAME;
		else
		{
			system->borderx=system->bordery=system_borderwidth;
			system->status|=SSTATUS_NOXBORDER|SSTATUS_NOYBORDER;
			system->win_xpos=(unsigned int)butret/65536;
			system->win_ypos=(short)butret;
			JOBInfo(MYSELF,NULL,NULL,&system->priority);
			JOBPriority(MYSELF,126);		/* magic priority for Button_Pick */
		}
	}

	/* determine position of window */
	system->curr_xsiz=xsiz;
	system->curr_ysiz=ysiz;

	return ERR_OK;
}

/* really display the window */
Error PWSetWindow(PWSystem *system)
{
	Error err;
	SubWindow *sw=system->subwin;
	int xsiz=system->curr_xsiz;
	int ysiz=system->curr_ysiz;
	Window win=system->window;

	/* determine position of window */
	system->xorg=-system->borderx;
	system->yorg=-system->bordery;
	PWDeterminePosition(system);

	/* create/draw the window */
	if (system->status&SSTATUS_BUTFRAME)
		catch( WINOutline(win, 0, 0,
					   xsiz+system->borderx*2, ysiz+system->bordery*2,
					   system->win_xpos, system->win_ypos) );
	else
		catch( WINOutline(win, system_xshadow, system_yshadow,
					   xsiz+system->borderx*2, ysiz+system->bordery*2,
					   system->win_xpos, system->win_ypos) );

	/* allocate a Gstate for this window -- as large as possible */
	catch( PFPGstateReAssign(system->gstate, PF_DRIVER_SCREEN, xsiz, ysiz,
			system->win_xpos+system->borderx,system->win_ypos+system->bordery) );

	/* now make sure origin of window is at 0,0 */
	PFPPageOriginSet(system->gstate,system->xorg+system->borderx,
									system->yorg+system->bordery);

	/* Draw move/resize border */
	PWDrawBorders(system);

	return err;
}
