/*
	ProWesS, PROGS Windowing System

	test program - file & directory select

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	started 26 August 1995 (based on test_files)
*/

#include "dir.h"
#include "dev.h"
#include "str.h"
#include "ctype.h"

#include "ProWesS.h"

typedef struct {
	PWObject dir,name,ext;
	PWObject all,tree,not;
	PWObject files, dirmenu;
	} FileGlobals;

/* forward declarations - for file select */
#define EXTENSIONLEN 128
Error getdir(PWObject object);
Error dirup(PWObject object);
Error fsel(PWObject object, char *item);
/* forward declarations - for directory select */
Error dirselect(PWObject object);
Error dirchange(PWObject object, char *item);

void init()
{
	FileGlobals g;
	PWObject obj,row,pref;
	PWObject sep1,sep2,sep3,sep4;
	PWObject dirbox, namebox, extbox;
	PWObject outl, devbox;
	char currdir[IO_MAXDIRECTORY];

	/* get current data directory to fill in directory name */
	DEVDataGet(currdir);

	/* create main window */
	PWCreate(NULL,&obj,PW_TYPE_OUTLINE,
				PW_OUTLINE_SLEEP,
				PW_OUTLINE_QUIT,
				PW_OUTLINE_ACTION_WAKE, &getdir,
				PW_OUTLINE_QUIT_KEYPRESS, 27,
				PW_WINDOW_YSIZE, short2pt(540),
				PW_GLOBAL_AUXILIARY, &g,
				NULL);

	/* create edline for filename, with label */
	PWCreate(obj,&namebox,PW_TYPE_DIRECTION,
				PW_SCALE_FACTOR, 0,
				NULL);
	PWCreate(namebox,NULL,PW_TYPE_LABEL,
				PW_LABEL_TEXT, "filename",
				PW_SCALE_FACTOR, 0,
				NULL);
	PWCreate(namebox,&g.name,PW_TYPE_EDLINE,
				PW_EDLINE_MAXLENGTH, IO_MAXFULLNAME,
				PW_EDLINE_KEYPRESS, 'f',
				NULL);

	/* create edline for filename, with label */
	PWCreate(obj,&extbox,PW_TYPE_DIRECTION,
				PW_SCALE_FACTOR, 0,
				NULL);
	PWCreate(extbox,NULL,PW_TYPE_LABEL,
				PW_LABEL_TEXT, "extensions",
				PW_SCALE_FACTOR, 0,
				NULL);
	PWCreate(extbox,&g.ext,PW_TYPE_EDLINE,
				PW_EDLINE_MAXLENGTH, EXTENSIONLEN,
				PW_EDLINE_KEYPRESS, 'e',
				PW_EDLINE_WIDTH_FS, short2pt(12),
				PW_EDLINE_ACTION_AFTER, &getdir,
				NULL);
	PWCreate(extbox,&g.not,PW_TYPE_LOOSE_ITEM,
				PW_LOOSE_TEXT, "Not",
				PW_LOOSE_KEYPRESS, 'n',
				PW_SCALE_FACTOR, 0,
				PW_LOOSE_ACTION_HIT, &getdir,
				NULL);

	/* create edline for directory, with label */
	PWCreate(obj,&dirbox,PW_TYPE_DIRECTION,
				PW_SCALE_FACTOR, 0,
				NULL);
	PWCreate(dirbox,NULL,PW_TYPE_LABEL,
				PW_LABEL_TEXT, "directory",
				PW_SCALE_FACTOR, 0,
				NULL);
	PWCreate(dirbox,&g.dir,PW_TYPE_EDLINE,
				PW_EDLINE_SET, currdir,
				PW_EDLINE_MAXLENGTH, IO_MAXDIRECTORY,
				PW_EDLINE_KEYPRESS, 'd',
				PW_EDLINE_ACTION_AFTER, &getdir,
				PW_EDLINE_WIDTH_FS, short2pt(12),
				PW_EDLINE_ACTION_DO, &dirselect,
				NULL);
	PWCreate(dirbox,&g.dirmenu,PW_TYPE_DIRSELECT,
				PW_KEYPRESS, 'D',
				PW_DIRSELECT_ACTION, &dirchange,
				NULL);
	PWCreate(dirbox,NULL,PW_TYPE_LOOSE_ITEM,
				PW_LOOSE_CHANGE_STATUS,FALSE,
				PW_LOOSE_ACTION_HIT, &dirup,
				PW_LOOSE_TEXT, "<-",
				PW_LOOSE_KEYPRESS, '<',
				PW_SCALE_FACTOR, 0,
				NULL);
	PWCreate(dirbox,&g.tree,PW_TYPE_LOOSE_ITEM,
				PW_LOOSE_TEXT, "Tree",
				PW_LOOSE_KEYPRESS, 't',
				PW_SCALE_FACTOR, 0,
				PW_LOOSE_ACTION_HIT, &getdir,
				NULL);
	PWCreate(dirbox,&g.all,PW_TYPE_LOOSE_ITEM,
				PW_LOOSE_TEXT, "All",
				PW_LOOSE_KEYPRESS, 'a',
				PW_SCALE_FACTOR, 0,
				NULL);

	/* create menu with possible filenames */
	PWCreate(obj,&sep4,PW_TYPE_SEPARATOR,
				NULL);
	PWCreate(obj,&g.files,PW_TYPE_MENU,
				PW_MENU_VISIBLE_LINES, 4,
				PW_MENU_ACTION_SELECT, &fsel,
				PW_MENU_SORT_COMPARE, &STRCompareCI,
				NULL);

	getdir(obj);

	/* activate */
	PWActivate(obj);
}

Error getdir(PWObject object)
{
	FileGlobals *g;
	DirIterator it;
	char currdir[IO_MAXDIRECTORY];
	char name[IO_MAXDIRECTORY+2];
	char ext[EXTENSIONLEN];
	int not,tree;
	FileInfo fi;

	PWQuery(object,PW_GLOBAL_AUXILIARY,&g);
	PWChange(g->dir,PW_EDLINE_GET,IO_MAXDIRECTORY,currdir,NULL);
	PWChange(g->dirmenu,PW_DIRSELECT_SET,currdir,NULL);
	PWChange(g->files,PW_MENU_CLEAR,NULL);
	PWChange(g->ext,PW_EDLINE_GET,EXTENSIONLEN,ext,NULL);
	PWQuery(g->not,PW_LOOSE_STATUS,&not);
	PWQuery(g->tree,PW_LOOSE_STATUS,&tree);
	not=(not==PW_STATUS_SELECTED);
	tree=(tree==PW_STATUS_SELECTED);

	/* get subdirectories */
	if (!tree && !DIRSubOpen(currdir,FALSE,&it))
	{
		while(!DIRSubNext(it,&fi))
		{
			STRCopy("> ",name);
			STRAppend(fi.name,name);
			PWChange(g->files,PW_MENU_ADD_COPY,name,NULL);
		}
		DIRSubClose(it);
	}

	/* get files in directory */
	if (!DIROpen(currdir,tree,&it,
				 not?DIRSieveExtensionNot:DIRSieveExtension,ext,NULL))
	{
		DIRDirectory(it,currdir);
		PWChange(g->dir,PW_EDLINE_SET,currdir,NULL);
		while(!DIRNext(it,&fi))
			PWChange(g->files,PW_MENU_ADD_COPY,fi.name,NULL);
		DIRClose(it);
	}

	return ERR_OK;
}

Error dirup(PWObject object)
{
	FileGlobals *g;
	char currdir[IO_MAXDIRECTORY];
	int len;

	PWQuery(object,PW_GLOBAL_AUXILIARY,&g);
	PWChange(g->dir,PW_EDLINE_GET,IO_MAXDIRECTORY,currdir,NULL);

	len=strlen(currdir);
	for (len=strlen(currdir)-1 ; len>0 && currdir[len-1]!='_' ; len--);
	if (len>0)
	{
		currdir[len]='\0';
		PWChange(g->dir,PW_EDLINE_SET,currdir,NULL);
		getdir(object);
	}
	return ERR_OK;
}

Error dirdown(PWObject object, char *item)
{
	FileGlobals *g;
	char currdir[IO_MAXDIRECTORY];

	PWQuery(object,PW_GLOBAL_AUXILIARY,&g);
	PWChange(g->dir,PW_EDLINE_GET,IO_MAXDIRECTORY,currdir,NULL);

	if (currdir[strlen(currdir)-1]!='_')
		STRAppendSafe("_",currdir,IO_MAXDIRECTORY);
	STRAppendSafe(item,currdir,IO_MAXDIRECTORY);
	PWChange(g->dir,PW_EDLINE_SET,currdir,NULL);
	getdir(object);

	return ERR_OK;
}

Error fsel(PWObject object, char *item)
{
	if (*item=='>')
		return dirdown(object,item+2);
	else
		return ERR_OK;
}

/* actions for directory select window */

Error dirselect(PWObject object)
{
	FileGlobals *g;
	char currdir[IO_MAXDIRECTORY];

	PWQuery(object,PW_GLOBAL_AUXILIARY,&g);

	return PWChange(g->dirmenu,PW_DIRSELECT_ACTIVATE,NULL);
}

Error dirchange(PWObject object, char *currdir)
{
	FileGlobals *g;

	PWQuery(object,PW_GLOBAL_AUXILIARY,&g);
	PWChange(g->dir,PW_EDLINE_SET,currdir,NULL);
	return getdir(object);
}
