/*
	ProWesS, PROGS Windowing System

	Application window object "APPL", a direction with in it a canvas and
		possibly a scrollbar.

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	started 5 May 1995
	10.04.1996 slight "optimization" in Define routine
	16.11.1997 PW_APPLIC_REDRAW added, redraws canvas & scroll bars
	16.01.1998 PW_APPLIC_CANVAS_CATCH added
*/

#include <mem.h>
#include <win.h>

#include "PWHandler.h"

#define catch(X) do { if (err=(X)) return err; } while (0)

/* global statusses */
char scrollleft=FALSE;
char scrollabove=FALSE;

/* object definition */
typedef struct {
	PWBaseObject;					/* common start for all objects */
	PWObject canvas;				/* canvas */
	PWObject dir, xbar, ybar;		/* for other objects */
} Object;

/* object handling routines */

Error Create(PWObject owner, PWObject *ret, PWEvent event, int *taglist)
{
	Error err;

	/* allocate specific object - which should be cleared (automatic) */
	catch( PWAllocate(event.region->system,sizeof(Object),ret) );

	/* no events to react to BUT is composite */
	event.region->composite=TRUE;

	return err;
}

Error Remove(PWObject object)
{
	/* all the other objects are children and are thus automatically removed */
	return PWRelease(object,object);
}

Error Change(PWObject object, int *taglist)
{
	Object *obj=(Object *)object;
	int op;
	PWSystem *system=obj->event.region->system;
	Error err=ERR_OK;
	int option;
	PWObject client;
	char unrecognised_tag=FALSE;

	if (!(obj->status&OSTATUS_INITIALISED))
	{
		obj->status|=OSTATUS_INITIALISED;
		/* create title, direction and sleep objects */
		catch( PWCreate((PWObject)obj, &obj->dir, PW_TYPE_DIRECTION, NULL) );
		catch( PWCreate(obj->dir, &obj->canvas, PW_TYPE_CANVAS, NULL) );
	}

	while (*taglist)
	{
		option=NUL; client=obj->canvas;

		switch (*taglist)
		{
		case PW_APPLIC_SCROLL_X:
			option=PW_SCROLL_NOBAR;
		case PW_APPLIC_SCROLLBAR_X:
			/* try to make the nesting of the regions minimal */
			{
				PWObject relpos, sep;
				if (obj->dir->event.region->level&1)
					relpos=obj->canvas;
				else
					relpos=obj->dir;
				catch( PWCreate((PWObject)obj,&sep,PW_TYPE_SEPARATOR,
							scrollabove?PW_POSITION_ABOVE:PW_POSITION_BELOW, relpos,
							NULL) );
				catch( PWCreate(sep,&obj->xbar,PW_TYPE_SCROLL,
							scrollabove?PW_POSITION_ABOVE:PW_POSITION_BELOW, sep,
							PW_SCROLL_CANVAS, obj->canvas,
							option,
							NULL) );
			}
			break;
		case PW_APPLIC_SCROLL_Y:
			option=PW_SCROLL_NOBAR;
		case PW_APPLIC_SCROLLBAR_Y:
			/* try to make the nesting of the regions minimal */
			{
				PWObject relpos, sep;
				if (!(obj->dir->event.region->level&1))
					relpos=obj->canvas;
				else
					relpos=obj->dir;
				catch( PWCreate((PWObject)obj,&sep,PW_TYPE_SEPARATOR,
							scrollleft?PW_POSITION_LEFT_OF:PW_POSITION_RIGHT_OF, relpos,
							NULL) );
				catch( PWCreate(sep,&obj->ybar,PW_TYPE_SCROLL,
							scrollleft?PW_POSITION_LEFT_OF:PW_POSITION_RIGHT_OF, sep,
							PW_SCROLL_CANVAS, obj->canvas,
							option,
							NULL) );
			}
			break;

		case PW_APPLIC_XSCROLL_LIST:
			client=obj->xbar;
			goto setlist;
		case PW_APPLIC_YSCROLL_LIST:
			client=obj->ybar;
		case PW_APPLIC_CANVAS_LIST:
setlist:	PWDoChange(client, taglist+1);
			break;

		case PW_APPLIC_REDRAW:
			obj->status|=OSTATUS_REDRAW;
			system->status|=SSTATUS_REDRAW;
			break;

		case PW_APPLIC_CANVAS_CATCH:
			PWChange((PWObject)obj,PW_CATCH_OBJECT,obj->canvas,NULL);
			break;

		default :
			if (*taglist&(~PW_TAGMASK_SKIP))
				unrecognised_tag=TRUE;
			break;
		}
		taglist=PWSkipTag(taglist);
	}
	if (unrecognised_tag) err=ERR_IPAR;
	return err;
}

Error Define(char *name, char *value)
{
	char *cname="APPLIC-SCROLLBAR-";
	while (*cname) if (*cname++!=*name++) return ERR_IPAR;

	if (STRSameCD(name,"LEFT"))
	{
		scrollleft=TRUE;
		return ERR_OK;
	}
	if (STRSameCD(name,"RIGHT"))
	{
		scrollleft=FALSE;
		return ERR_OK;
	}
	if (STRSameCD(name,"ABOVE"))
	{
		scrollabove=TRUE;
		return ERR_OK;
	}
	if (STRSameCD(name,"BELOW"))
	{
		scrollabove=FALSE;
		return ERR_OK;
	}
	return ERR_IPAR;
}

PWTypeDef init={
	PW_TYPEDEF_FLAG, PW_TYPE_APPLIC, PW_TYPE_EVENT_REGION,
	&Create, &Change, NULL, &Remove, NULL, NULL, NULL, &Define,
	NULL, NULL, NULL
};
