/*
	ProWesS, PROGS Windowing System

	item support type "iTEM"

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	started 26 January 1995
	18.12.1995 added Draw routine to have equivalent type without active border
	21.03.1996 changed internal data to more pixels
	06.06.1996 use system default ink when none defined
	28.06.1996 faster (and shorter) border drawing thanks to PFP and better
			   handling of linewidth in PROforma
	17.01.1997 adjusted to use part of window PFPageShow -> faster
*/

#include <mem.h>

#include "PWHandler.h"

/* global statusses */

#define p100 short2pt(100)
pt Xpt2pix, Ypt2pix;
pt width=0;
pix xwidth=1,ywidth=1;		/* width nominal and each direction */
ColourRGB ink	= { p100, 0, 0 }; char Kink  =FALSE;
ColourRGB paper = { 0,    0, 0 }; char Kpaper=FALSE;

static Error DrawBorder(PWObject obj, ColourRGB *colour)
{
	PWRegion *reg=obj->event.region;
	Gstate gstate=reg->system->gstate;
	pix xorg=reg->hit_xorg-xwidth;
	pix yorg=reg->hit_yorg-ywidth;
	pix xsiz=reg->hit_xend-xorg;
	pix ysiz=reg->hit_yend-yorg;

	PFPathStroked(gstate);
	PFColourRGB(gstate, colour);
	PFLineWidth(gstate, width);

	PFPWindowSubSet(gstate,xsiz+xwidth,ysiz+ywidth,xorg,yorg);
	PFPMoveTo(gstate, short2pt(xwidth)/2, short2pt(ywidth)/2);
	PFPRectangle(gstate, short2pt(xsiz),short2pt(ysiz));
	PFPathDraw(gstate);
	PFPageShow(gstate);
	return PFWindowSubRestore(gstate);
}

Error Entry(PWObject object)
{
	return DrawBorder(object,&ink);
}

Error Exit(PWObject object)
{
	return DrawBorder(object,&paper);
}

Error Draw(PWObject objetc)
{
	return ERR_OK;
}

Error Define(char *name, char *value)
{
	ColourRGB *colour;
	double red, green, blue, db;

	LinkPROforma();
	PWpoint2pixel(&Xpt2pix,&Ypt2pix);

	if (STRSameCD(name,"ITEM-BORDER-COLOUR"))
	{
		Kink=TRUE;
		colour=&ink;
readcolour:
		STRFormatGet(value,NULL,"%f %f %f",&red,&green,&blue);
		colour->red  =double2pt(red);
		colour->green=double2pt(green);
		colour->blue =double2pt(blue);
		return ERR_OK;
	}
	if (STRSameCD(name,"SYSTEM-BACKGROUND-COLOUR"))
	{
		Kpaper=TRUE;
		colour=&paper;
		goto readcolour;
	}
	if (STRSameCD(name,"ITEM-BORDER-WIDTH"))
	{
		STRFormatGet(value,NULL,"%f",&db);
		width=double2pt(db);

		xwidth=pt2rshort(fixmul(width,Xpt2pix)); if (xwidth<1) xwidth=1;
		ywidth=pt2rshort(fixmul(width,Ypt2pix)); if (ywidth<1) ywidth=1;
		return ERR_OK;
	}
	return ERR_IPAR;
}

Error DetermineSize(PWObject obj)
{
	PWRegion *region=obj->event.region;

	if (!Kink)	 PWQuery(obj,PW_DEFAULT_MIDDLEGROUND,&ink);
	if (!Kpaper) PWQuery(obj,PW_DEFAULT_BACKGROUND,&paper);

	region->hit_off_lft=region->hit_off_rgt=xwidth;
	region->hit_off_top=region->hit_off_bot=ywidth;

	region->tot_xsiz+=2*short2pt(xwidth);
	region->tot_ysiz+=2*short2pt(ywidth);
	return ERR_OK;
}

PWTypeDef init={
	PW_TYPEDEF_FLAG, PW_TYPE_iTEM, PW_TYPE_EVENT_REGION,
	NULL, NULL, NULL, NULL, Entry, Exit, NULL, Define,
	Draw, DetermineSize, NULL
};
