/*
	ProWesS, PROGS Windowing System

	Application window object "MENU", a future replacement for menu
	closely copied on applic, canvas and menu....

	written by Wolfgang Lenerz based on applic, notice for original "MENU"
	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	started 16 July 1995
	14.10.1995 made sure string have a pixel at either side in menu
	23.03.1996 adjusted for internal use of pixel values
	10.04.1996 use PWHandleDefine && use ProWesS default font & fontsize
	02.06.1996 PW_MENU_ADD and variants now set the current item to allow the
			   user to change the status using PW_MENU_STATUS_CURRENT
	06.06.1996 - can now use default colours when none given explicitly
			   - added PW_MENU_KEYPRESSES create tag and use provide standard
	11.06.1996 (thanks to Wolfgang Lenerz) added PW_MENU_FIRST_NUMBER,
			   PW_MENU_NEXT_NUMBER, PW_MENU_AVAILABLE_FIRST_NUMBER,
			   PW_MENU_AVAILABLE_NEXT_NUMBER, PW_MENU_SELECTED_FIRST_NUMBER,
			   PW_MENU_SELECTED_NEXT_NUMBER, PW_MENU_STATUS_ALL
	01.07.1996 PW_MENU_STATUS was bugged when PW_MENU_UNIQUE
	17.09.1996 PW_MENU_UNIQUE now also maintained when setting a status
			   explicitly
	18.09.1996 all PW_MENU_ITEMWIDTH_xxx were interpretted as MAX
	24.09.1996 PW_MENU_ITEMWIDTH_COLUMNS added
	27.09.1996 another attempt at fixing PW_MENU_UNIQUE bug (thanks Wolfgang)
	07.10.1996 added PW_MENU_SELECTED_NUMBER, PW_MENU_AVAILABLE_NUMBER and
			   PW_MENU_UNAVAILABLE_NUMBER (wl)
	25.11.1996 item height now better
	07.12.1996 use PFPShow/Width/FontBbox
	19.12.1996 PW_MENU_STATUS_CURRENT query added

	26.12.1996 (wl) - removed dependence onother types
					- indicating several items by dragging introduced
					- only the newly (de)selected item redrawn
			   However, this does change the direction !
	16.01.1996 - redraw revamped
			   - when PW_MENU_UNIQUE the deselected item also redrawn
			   - removed some global vars used by wl
			   - direction works properly again (needs PWDoCreate)
	17.01.1996 - when redrawing the entire menu, no PFPageShow for each item
			   - DrawBorder uses the fact that PFPageShow only redraws the area
				 inside the WindowSub for faster redraw
	19.01.1997 border of last current item also undrawn when moving into empty
			   space where no more menu items are
	27.01.1997 - PW_MENU_BASIC_TRUEMENU
			   - really undraws last item border, and also redraws it if
				 pointer goes in again
	06.02.1997 PW_MENU_CLEAR also sets obj->selected to NULL and resets
			   row, column and current to -1 (solves several pbs with border
			   drawing for current items)
	20.04.1997 autosizing of menus to match contents
	23.04.1997 autosizing now also automatic when room and adding items
	05.05.1997 - added an Entry routine which draws a border when necessary
			   - PW_MENU_ACTION_DOSELECT added
	07.05.1997 autosizing now also shrinks menu horizontally when necessary
	13.05.1997 menu items now drawn in more natural order (by column)
	21.05.1997 - PW_MENU_SORT_COMPARE now properly redraws window
			   - menu no longer displays border around partially visible lines
				 which contain nothing
	27.05.1997 WINDOW_DO still creates a window do when not in an object and
			   there is a DoSelect routine (bugged since DoSelect introduced)
	06.06.1997 New define DISPLAY-ROWS allows menu to be drawn in the old way
			   (by rows) (wl)
	16.06.1997 DoSelect routine still called when DO on selected item, this
			   allows windowdo to occur
*/

#include <mem.h>
#include <win.h>
#include "PWHandler.h"

#define catch(X) do { if (err=(X)) return err; } while (0)

typedef struct { ColourRGB ink, paper; } ColourStatus;

/* global statusses for applic (scroll) */
char scrollleft =FALSE;
char scrollabove=FALSE;

/* global statusses for "canvas"*/
pt Xpix2pt, Ypix2pt;
pt Xpt2pix, Ypt2pix;

/* global statusses for "menu"*/
#define STATUS_MAXWIDTH (short)0x0001
#define STATUS_COLUMNS	(short)0x0002
#define STATUS_UNIQUE	(short)0x0004
#define STATUS_NOSELECT (short)0x0008
#define STATUS_WINDOWDO (short)0x0010
#define p100 short2pt(100)
#define p1 short2pt(1)
ColourRGB ink	= { 0,    0,    0    }; char Kink=   FALSE;
ColourRGB paper = { p100, p100, p100 }; char Kpaper= FALSE;
ColourRGB border= { p100, 0,    0    }; char Kborder=FALSE;
ColourStatus selected	={{ p100, p100, p100 },{ 0   , 0   , 0    }};
ColourStatus available	={{ 0	, 0   , 0  },{ p100, p100, p100 }};
ColourStatus unavailable={{ 0 , p100, 0  },{ p100, p100, p100 }};
char font[PF_MAXFONTNAME]="";	/* name of PROforma font to use */
pt fontsize=0;						/* fontsize to use */
pt baseline=0;						/* font baseline position (at fontsize) */
pix height=0, itemheight;			/* height of font (at fontsize) */
pix xwidth=0, ywidth=0; pt width;	/* width nominal and in each direction */
char recalculate=TRUE;

char disp_rows=FALSE;				/* display in rows? default = no */

/* loose item object definition */
typedef struct _item {
	struct _item *next; 	/* next and .. */
	struct _item *prev; 	/* .. previous item in menu */
	char *item; 	/* string for this item */
	char status;			/* selected, available or unavailable ? */
	char copied;			/* was the item copied ? */
} Item;

/* object definition */
typedef struct {
  /* from canvas */
	PWBaseObject;			/* common start for all objects */
	pix winwidth;			/* real width of window */
	pix winxsiz;			/* (min) size of "canvas" on screen */
	pix winysiz;
	pix winxorg;			/* origin of "canvas" on screen */
	pix winyorg;
  /* from applic */
	PWObject dir, scroll;	/* for other objects */
  /* from menu */
	Item *base, *sentinel;	/* base (before) and sentinel item */
	int count;				/* number of items in menu */
	int inrow;				/* no of items displayed side by side */
	Error (*Select)(PWObject,char *);	/* select action routine */
	Error (*DoSelect)(PWObject,char *); /* select action routine when using DO */
	Error (*DeSelect)(PWObject,char *); /* deselect action routine */
	pix itemwidth;			/* requested width for items */
	pix realwidth;			/* real current itemwidth */
	short mstatus;			/* several boolean statusses */
	Item *selected; 		/* currently selected object when STATUS_UNIQUE */
	int current;			/* count of item currently pointed */
	Error (*sortcmp)(char *s1, char *s2, int *res); /* cmp rout for sorting */
	Item *qcurrent; 		/* current item for query operators */
	char ksu, ksd, klu, kld;/* keys for scrollbar */
	int row;				/* row of current object */
	int column; 			/* column of current object */
	short drag_is_on;		/* if set, this shows that we are dragging */
	int lines_max;			/* max no of lines which are possible in menu */
	int lines_have; 		/* number of lines in menu */
	int lines_vis;			/* minimum number of visible lines */
	int lines_inwin;		/* number of lines which are visible in window */
} Object;

/* -------------------------------------------- */
/* First some sub-routines */
static void Calculate(Gstate gstate, PWObject object)
{
	PBbox bbox;

	if (!Kpaper) PWQuery(object,PW_DEFAULT_BACKGROUND,&paper);
	if (!Kink) PWQuery(object,PW_DEFAULT_FOREGROUND,&ink);
	if (!Kborder) PWQuery(object,PW_DEFAULT_MIDDLEGROUND,&border);
	if (!fontsize) PWQuery(object,PW_DEFAULT_FONTSIZE,&fontsize);
	if (!*font) PWQuery(object,PW_DEFAULT_FONT,font);
	PFFontLoad(gstate,font);
	PFFontScale(gstate,fontsize);
	PFPFontBbox(gstate,'_',&bbox);
	baseline=short2pt(bbox.yorg);
	height=bbox.ysiz;			/* height text in item */
	itemheight=height+ywidth;	/* height of item + border */
	recalculate=FALSE;
}

/* determine number of visible items and realwidth */
Error menu_calc(Object *obj)
{
	int inrow;
	if (recalculate) Calculate(obj->event.region->system->gstate,(PWObject)obj);
	 
	if (obj->mstatus&STATUS_COLUMNS)
	{
		inrow=obj->itemwidth;
	}
	else
	{
		inrow=(obj->winwidth-xwidth)/(obj->itemwidth+xwidth);
		if (obj->count==0) inrow=1;
	}
	if (inrow<1) inrow=1;
	obj->realwidth=(obj->winwidth-xwidth)/inrow-xwidth;

	if (inrow!=obj->inrow)
	{
		obj->inrow=inrow;
		obj->lines_have=(obj->count+inrow-1)/inrow;
		PWChange(obj->scroll,
					PW_SCROLL_MAXIMUM, obj->lines_have,
					PW_SCROLL_SCROLL, 0,
					NULL);
	}
	return ERR_OK;
}
/* routine to change the itemwidth */
Error change_width(Object *obj, short width)
{
	PWSystem *system=obj->event.region->system;
	width+=2;

	if (width>obj->itemwidth)
	{
		obj->itemwidth=width;
		if (width+xwidth*2 > obj->winxsiz)
		{
			obj->winxsiz=width+xwidth*2;
			system->status|=SSTATUS_REDRAW;
			obj->status|=OSTATUS_REDRAW;
		}
		else
			menu_calc(obj);
	}
	return ERR_OK;
}

/* this puts item in sorted list */
Error sort_add(Object *obj, Item *item)
{
	Item *curr=obj->base->next;
	int res;

	while(curr->next)
	{
		obj->sortcmp(curr->item,item->item,&res);
		if (res>0)
			break;
		else
			curr=curr->next;
	}
	item->prev=curr->prev;
	item->next=curr;
	curr->prev->next=item;
	curr->prev=item;
	return ERR_OK;
}

/* routine to add a string in the object */
Error add_string(Object *obj, char *str, char copy)
{
	PWSystem *system=obj->event.region->system;
	Gstate gstate=system->gstate;
	Item *new;
	Error err;
	int lines_have;

	/* allocate and fill in */
	catch( PWAllocate(system,sizeof(Item),&new) );
	new->status=PW_STATUS_AVAILABLE;
	new->copied=copy;
	if (copy)
	{
		err=PWAllocate(system,strlen(str)+1,&new->item);
		if (err) { PWRelease((PWObject)obj,new); return err; }
		STRCopy(str,new->item);
	}
	else
		new->item=str;

	/* link in */
	if (!obj->sortcmp)
	{
		new->next=obj->sentinel;
		new->prev=obj->sentinel->prev;
		obj->sentinel->prev->next=new;
		obj->sentinel->prev=new;
	}
	else
	{
		sort_add(obj,new);
	}
	obj->qcurrent=new;				/* newly added item should be current */

	/* increase number of items in applic */
	lines_have=((obj->count++)+obj->inrow)/obj->inrow;
	if (lines_have!=obj->lines_have)
	{
		PWChange(obj->scroll,
					PW_SCROLL_MAXIMUM, lines_have,
					NULL);
		obj->lines_have=lines_have;
		if (obj->lines_max>=lines_have)
			system->status|=SSTATUS_WINCHANGE;
	}
	system->status|=SSTATUS_REDRAW;
	obj->status|=OSTATUS_REDRAW; /* this must be redrawn */

	/* set new width for items */
	if (obj->mstatus&STATUS_MAXWIDTH)
	{
		short width;
		PFFontLoad(gstate,font);
		PFFontScale(gstate,fontsize);
		PFPStringWidth(gstate,str,&width);
		change_width(obj,width);
	}
	return err;
}

/* ---------------------------------------- */
/* event handler routine for scrollbar inside menu */
/* calculate size of bar */
Error calcsize(PWObject object, pt *size)
{
	if (recalculate) Calculate(object->event.region->system->gstate,object);
	*size=(PF2ypix(*size)+1-ywidth)/itemheight;
	return ERR_OK;
}

/*----------------------------------------*/
/* object handling routines */

Error Remove(PWObject object)
{
	/* all the other objects are children and are thus automatically removed */
	return PWRelease(object,object);
}
 
Error Create(PWObject owner, PWObject *ret, PWEvent event, int *taglist)
{
	Error err=ERR_OK;
	Object *obj;
	PWSystem *system=event.region->system;	/* this is my system */
	LinkPROforma();
	PWpixel2point(&Xpix2pt,&Ypix2pt);
	PWpoint2pixel(&Xpt2pix,&Ypt2pix);


	/* allocate specific object - which should be cleared (automatic) */
	catch (PWAllocate(event.region->system,sizeof(Object),ret));

	obj=(Object *)*ret;
	if (recalculate) Calculate(event.region->system->gstate, (PWObject)obj);

	event.region->composite=TRUE;
	event.region->wait_event=PW_EVENT_HIT|PW_EVENT_DO|PW_EVENT_MOVE
							 |PW_EVENT_HITDRAG; /* events to react on */

	obj->mstatus=STATUS_MAXWIDTH; /* use maximum width of items */
	obj->inrow=1;					/* there is one item in each row */
	obj->lines_vis=1;				/* at least one visible line */
	obj->current=-1;				/* no current item yet */
	obj->row=obj->column=-1;

	/* set standard keypresses for scrolling the menu */
	obj->ksu=(char)209; obj->ksd=(char)217;
	obj->klu=(char)213; obj->kld=(char)221;

	/* initialise list of items (base and sentinel) */
	err=PWAllocate(system,sizeof(Item),&obj->base);
	if (err) { Remove((PWObject)obj); return err; }
	err=PWAllocate(system,sizeof(Item),&obj->sentinel);
	if (err) { Remove((PWObject)obj); return err; }
	obj->base->next=obj->sentinel;
	obj->sentinel->prev=obj->base;

	/* see if there is a PW_MENU_KEYPRESSES tag */
	while (*taglist)
	{
		if (*taglist==PW_MENU_KEYPRESSES)
		{
			/* set different keypresses for scrolling the menu */
			obj->ksu=(*(taglist+1)); obj->ksd=(*(taglist+2));
			obj->klu=(*(taglist+3)); obj->kld=(*(taglist+4));
			*taglist&=PW_TAGMASK_SKIP;
		}
		taglist=PWSkipTag(taglist);
	}
	return err;
}

/* redraw one item */
Error redraw_item(Object *obj, Item *item, int row, int column, char show)
{
	PWRegion *reg=obj->event.region;
	PWSystem *system=reg->system;
	Gstate gstate=system->gstate;
	ColourStatus *status;

	if (row==-1 || column==-1)
	{
		/* see where this item is */
		int i; Item *list=obj->base->next;
		for (i=0 ; list!=item ; i++,list=list->next);
		row=i%obj->lines_have-obj->winyorg;
		column=i/obj->lines_have;
	}
	if (row<0 || row>obj->lines_inwin) return ERR_OK;

	status=&selected;
	if (item->status==PW_STATUS_UNAVAILABLE) status=&unavailable;
	if (item->status==PW_STATUS_AVAILABLE) status=&available;
	/* draw item */
	PFPWindowSubSet(gstate,obj->realwidth,height,
						  column*(obj->realwidth+xwidth)+xwidth+reg->hit_xorg,
						  row	*	   itemheight		+ywidth+reg->hit_yorg);
	PFPaperColourRGB(gstate,&status->paper);
	PFColourRGB(gstate,&status->ink);
	PFPageClear(gstate);
	PFFontLoad(gstate,font);
	PFFontScale(gstate,fontsize);
	PFPMoveTo(gstate,pt_one,baseline);
	PFPStringShow(gstate,item->item);
	if (show) PFPageShow(gstate);	  /* redraw this now */
	PFWindowSubRestore(gstate);
	return ERR_OK;
}
 
/* this Draws the entire object */
Error Draw(PWObject object)
{
	Error err=ERR_OK;
	Object *obj=(Object *)object;
	PWRegion *reg=obj->event.region;
	PWSystem *system=reg->system;
	Gstate gstate=system->gstate;
	int count, my_row, my_col;
	Item *item=obj->base->next;
	pix ypixsiz=reg->hit_yend-reg->hit_yorg;
	if (recalculate) Calculate(gstate,object);

	/* clear drawing area and set some drawing parameters */
	PFPWindowSubSet(gstate, reg->hit_xend-reg->hit_xorg,
							reg->hit_yend-reg->hit_yorg,
							reg->hit_xorg,
							reg->hit_yorg);
	PFPaperColourRGB(gstate,&paper);
	PFPageClear(gstate);
	PFWindowSubRestore(gstate);

	/* display all items until the menu is filled */
	if (disp_rows)
	{           /* display in rows */

		/* search for first item to display */
		for (item=obj->base->next,count=obj->winyorg*obj->inrow ;
			 item->next && count ;
			 item=item->next,count--);
		for (my_row=my_col=0 ; item->next ; item=item->next)
		{
			redraw_item(obj,item,my_row,my_col,FALSE);

			/* move to next position */
			my_col++;
			if (my_col==obj->inrow)
			{
				my_row++; my_col=0;
				if (my_row*itemheight+ywidth>ypixsiz) break;
			}
		}
	}
	else
	{           /* display in columns */
		for (my_row=my_col=0 ; item->next ; item=item->next)
		{
			redraw_item(obj,item,my_row-obj->winyorg,my_col,FALSE);
			/* move to next position */
			my_row++;
			if (my_row==obj->lines_have)
			{
				my_col++; my_row=0;
			}
		}
	}
	system->status|=SSTATUS_PAGESHOW;
	return ERR_OK;
}

/* all possible changes for the object */
Error Change(PWObject object, int *taglist)
{
	Object *obj=(Object *)object;
	int op;
	PWSystem *system=obj->event.region->system;
	Error err=ERR_OK;
	int option;
	char unrecognised_tag=FALSE;
	Gstate gstate=system->gstate;

	if (!(obj->status&OSTATUS_INITIALISED))
	{
		PWObject relpos, sep;
		obj->status|=OSTATUS_INITIALISED;		/* init is done */

		/* set initial size now */
		system->status|=SSTATUS_WINCHANGE;		/* redraw system after this */
		obj->winysiz=obj->lines_vis*itemheight+ywidth;	/* new y size */

		/* create separator and scroll bar objects */
		catch( PWCreate((PWObject)obj,&sep,PW_TYPE_SEPARATOR,
					scrollleft?PW_POSITION_LEFT_OF:PW_POSITION_RIGHT_OF, object,
					NULL) );
		catch( PWCreate(sep,&obj->scroll,PW_TYPE_SCROLL,
					scrollleft?PW_POSITION_LEFT_OF:PW_POSITION_RIGHT_OF, sep,
					PW_OBJECT_AUXILIARY, obj,
					PW_SCROLL_CANVAS, obj,		/* pretend we're a canvas */
					PW_SCROLL_MINDIST, 1,
					PW_SCROLL_MAXDIST, -1,
					PW_SCROLL_CALCSIZE, &calcsize,
					PW_SCROLL_MIN_KEYPRESSES, obj->ksu, obj->ksd,
					PW_SCROLL_MAX_KEYPRESSES, obj->klu, obj->kld,
					NULL) );
	}
	while (*taglist)	/* do this until no more params */
	{
		switch (*taglist)
		{
/* allow redrawing of the 'canvas' from the scroll object */
		case PW_CANVAS_YSCROLL:
			obj->winyorg+=*(taglist+1);
		case PW_CANVAS_REDRAW:
			system->status|=SSTATUS_REDRAW;
			obj->status|=OSTATUS_REDRAW;	/* just set redraw */
			break;

/* now MENU changes */
		case PW_MENU_ADD:
			catch( add_string(obj,(char *)(*(taglist+1)),FALSE) );
			break;
		case PW_MENU_ADD_COPY:
			catch( add_string(obj,(char *)(*(taglist+1)),TRUE) );
			break;
		case PW_MENU_ADD_ARRAY:
			{
				int i,count=(*(taglist+1));
				int length=(*(taglist+2));
				char *str=(char *)(*(taglist+3));
				for (i=count ; i ; i--, str+=length)
					catch( add_string(obj,str,FALSE) );
			}
			break;
		case PW_MENU_STATUS_CURRENT:
			if (obj->mstatus&STATUS_NOSELECT &&
				(*(taglist+1))==PW_STATUS_SELECTED)
				break;
			if (obj->qcurrent)
				obj->qcurrent->status=(*(taglist+1));
			if (obj->mstatus&STATUS_UNIQUE &&
				(*(taglist+1))==PW_STATUS_SELECTED &&
				obj->selected!=obj->qcurrent)
			{
				obj->selected->status=PW_STATUS_AVAILABLE;
				if (obj->DeSelect)
					obj->DeSelect((PWObject)obj,obj->selected->item);
				obj->selected=obj->qcurrent;
			}
			system->status|=SSTATUS_REDRAW;
			obj->status|=OSTATUS_REDRAW;
			break;
		case PW_MENU_STATUS:
			if (obj->mstatus&STATUS_NOSELECT &&
				(*(taglist+1))==PW_STATUS_SELECTED)
				break;
			{
				Item *item=obj->base->next;
				while (item->next && item->item!=(char *)(*(taglist+2)))
					item=item->next;
				if (item->item==(char *)(*(taglist+2)))
					item->status=(*(taglist+1));
				if (obj->mstatus&STATUS_UNIQUE &&
					(*(taglist+1))==PW_STATUS_SELECTED &&
					obj->selected && obj->selected!=item)
				{
					obj->selected->status=PW_STATUS_AVAILABLE;
					if (obj->DeSelect)
						obj->DeSelect((PWObject)obj,obj->selected->item);
				}
				if ((*(taglist+1))==PW_STATUS_SELECTED)
					obj->selected=item;

				system->status|=SSTATUS_REDRAW;
				obj->status|=OSTATUS_REDRAW;
			}
			break;
		case PW_MENU_STATUS_ALL:
			/* can't select all if not all selectable */
			if (obj->mstatus&(STATUS_UNIQUE|STATUS_NOSELECT))
				break;
			{
				Item *item=obj->base->next;
				while (item->next)
				{
					item->status=(*(taglist+1));
					item=item->next;
				}
			}
			system->status|=SSTATUS_REDRAW;
			obj->status|=OSTATUS_REDRAW;
			break;
		case PW_MENU_ITEMWIDTH:
			obj->itemwidth=PF2xpix(*(taglist+1));
			goto setitemwidth;
		case PW_MENU_ITEMWIDTH_PIX:
			obj->itemwidth=(*(taglist+1));
			goto setitemwidth;
		case PW_MENU_ITEMWIDTH_FS:
			obj->itemwidth=PF2xpix(fixmul((*(taglist+1)),fontsize));
setitemwidth:
			obj->mstatus&=~(STATUS_MAXWIDTH+STATUS_COLUMNS);
			system->status|=SSTATUS_WINCHANGE;
			obj->winxsiz=obj->itemwidth+xwidth*2;
			menu_calc(obj);
			break;
		case PW_MENU_ITEMWIDTH_COLUMNS:
			obj->mstatus|=STATUS_COLUMNS;
			obj->mstatus&=~STATUS_MAXWIDTH;
			obj->itemwidth=(*(taglist+1));
			menu_calc(obj);
			break;
		case PW_MENU_ITEMWIDTH_MAX:
			obj->mstatus&=~STATUS_COLUMNS;
			obj->mstatus|=STATUS_MAXWIDTH;
			obj->itemwidth=0;
			{
				Item *list=obj->base->next;
				short width;
				PFFontLoad(gstate,font);
				PFFontScale(gstate,fontsize);
				for (;list->next;list=list->next)
				{
					PFPStringWidth(gstate,list->item,&width);
					change_width(obj,width);
				}
			}
			break;
		case PW_MENU_ACTION_SELECT:
			obj->Select=(Error (*)(PWObject,char *))(*(taglist+1));
			break;
		case PW_MENU_ACTION_DOSELECT:
			obj->DoSelect=(Error (*)(PWObject,char *))(*(taglist+1));
			break;
		case PW_MENU_ACTION_DESELECT:
			obj->DeSelect=(Error (*)(PWObject,char *))(*(taglist+1));
			break;
		case PW_MENU_NONE_SELECTED:
			obj->mstatus|=STATUS_NOSELECT;
		case PW_MENU_UNIQUE:
			obj->mstatus|=STATUS_UNIQUE;
			obj->selected=NULL; 	/* no currently selected item */
			/* reset status of all objects to available */
			{
				Item *list=obj->base->next;
				for (;list->next;list=list->next)
					list->status=PW_STATUS_AVAILABLE;
			}
			system->status|=SSTATUS_REDRAW;
			obj->status|=OSTATUS_REDRAW;
			break;
		case PW_MENU_SORT_COMPARE:
			obj->sortcmp=(Error (*)(char *,char *,int *))(*(taglist+1));
			if (obj->sortcmp && obj->base->next->next)
			{
				/* resort the list with compare routine */
				Item *list=obj->base->next;

				obj->sentinel->prev->next=NULL;
				obj->base->next=obj->sentinel;
				obj->sentinel->prev=obj->base;

				while (list)
				{
					Item *curr=list;
					list=list->next;
					sort_add(obj,curr);
				}
			}
			system->status|=SSTATUS_REDRAW;
			obj->status|=OSTATUS_REDRAW;
			break;
		case PW_MENU_VISIBLE_LINES:
			system->status|=SSTATUS_WINCHANGE;
			obj->lines_vis=max(1,(*(taglist+1)));
			obj->winysiz=obj->lines_vis*itemheight+ywidth;
			break;
		case PW_MENU_CLEAR:
			/* release all allocated strings and the list */
			{
				Item *it, *base=obj->base;
				while (base->next->next)
				{
					it=base->next;
					if (it->copied)
						PWRelease((PWObject)obj,it->item);
					base->next=it->next;
					PWRelease((PWObject)obj,it);
				}
			}
			obj->sentinel->prev=obj->base;
			obj->count=0;
			obj->qcurrent=obj->selected=NULL;
			obj->row=obj->column=obj->current=-1;
			obj->lines_have=0;
			/* reset number of items in scroll */
			PWChange(obj->scroll,
							PW_SCROLL_MAXIMUM, 0,
							PW_SCROLL_SCROLL, 0,
							NULL);
			/* reset itemwidth ?? */
			if (obj->mstatus&STATUS_MAXWIDTH)
				obj->itemwidth=0;

			system->status|=SSTATUS_REDRAW;
			obj->status|=OSTATUS_REDRAW;

			/* reset window size if autosizing */
			if (obj->lines_max)
			{
				obj->winysiz=obj->lines_vis*itemheight+ywidth;
				PWChange((PWObject)obj,PW_WINDOW_YSIZE,0,NULL);
			}
			break;
		case PW_MENU_WINDOW_DO:
			if (*(taglist+1))
				obj->mstatus|=STATUS_WINDOWDO;
			else
				obj->mstatus&=~STATUS_WINDOWDO;
			break;
		default :
			if (*taglist&(~PW_TAGMASK_SKIP))
				unrecognised_tag=TRUE;
			break;
		}
		taglist=PWSkipTag(taglist);
	}
	if (unrecognised_tag) err=ERR_IPAR;
	return err;
}

static Error DrawBorder(Object *obj,ColourRGB *colour)
{
	PWRegion *reg=obj->event.region;
	PWSystem *system=reg->system;
	Gstate gstate=system->gstate;
	pix xsiz=obj->realwidth+xwidth;

	if (obj->current<0 || obj->row<0 || obj->column<0) return ERR_OK;

	if (recalculate) Calculate(gstate,(PWObject)obj);

	PFPathStroked(gstate);
	PFColourRGB(gstate, colour);

	PFPWindowSubSet(gstate,xsiz+xwidth,itemheight+ywidth,
						   xsiz 		* obj->column+reg->hit_xorg,
						   itemheight	* obj->row	 +reg->hit_yorg);
	PFLineWidth(gstate,width);
	PFPMoveTo(gstate,short2pt(xwidth)/2,short2pt(ywidth)/2);
	PFPRectangle(gstate, short2pt(xsiz), short2pt(itemheight));
	PFPathDraw(gstate);
	PFPageShow(gstate);
	PFWindowSubRestore(gstate);
	return ERR_OK;
}

/* routine called when pointer leaves region - should undraw current item border*/
Error Exit(PWObject object)
{
	Object *obj=(Object *)object;
	DrawBorder(obj,&paper);
	obj->current=obj->row=obj->column=-1;
	obj->drag_is_on=0;				/* we aren't dragging anymore */
	return ERR_OK;
}

/* called each time an item is hit or done */
Error menu_hit_or_done(Object *obj, PWWait event)
{
	Item *list;
	int i;
	Error err=ERR_OK;
	int current=obj->current;
	if (current<0) goto exit_key;

	for (i=current,list=obj->base->next ; i ; i--, list=list->next);
	if (!(obj->mstatus&STATUS_NOSELECT))
	{
		if (event==PW_EVENT_HIT)
		{
			if (list->status==PW_STATUS_UNAVAILABLE) goto exit_key;
			list->status^=PW_STATUS_SELECTED;
			if (list->status==PW_STATUS_AVAILABLE && obj->DeSelect)
				err=obj->DeSelect((PWObject)obj,list->item);
		}
		else
		{
			if (list->status==PW_STATUS_UNAVAILABLE)
				{ current=-1; goto exit_key; }
			if (list->status==PW_STATUS_SELECTED) goto callDoSel;
			list->status=PW_STATUS_SELECTED;
		}
		if (list->status==PW_STATUS_SELECTED && obj->mstatus&STATUS_UNIQUE)
		{
			if (obj->selected && obj->selected!=list)
			{
				obj->selected->status=PW_STATUS_AVAILABLE;
				if (obj->DeSelect)
					err=obj->DeSelect((PWObject)obj,obj->selected->item);
				redraw_item(obj,obj->selected,-1,-1,TRUE);
			}
			obj->selected=list;
		}
		redraw_item(obj,list,obj->row,obj->column,TRUE);
	}

	if (list->status==PW_STATUS_SELECTED || obj->mstatus&STATUS_NOSELECT)
	{
callDoSel:
		if (event==PW_EVENT_DO && obj->DoSelect)
			err=obj->DoSelect((PWObject)obj,list->item);
		else
			if (obj->Select)
				err=obj->Select((PWObject)obj,list->item);
	}
exit_key:
	if (event==PW_EVENT_DO && obj->mstatus&STATUS_WINDOWDO &&
		( !obj->DoSelect || current<0 ) )
		PWHandleKeyPress(obj->event.region->system);
	return err;
}

/* called each time pointer moves inside object */
Error pointer_move(Object *obj)
{
	PWRegion *reg=obj->event.region;
	PWSystem *system=reg->system;
	int current=-1; 		/* current item - none just yet */
	int new_row,new_column;
	pix ptrxpos=system->ptr_xpos-system->win_xpos-reg->hit_xorg+system->xorg;
	pix ptrypos=system->ptr_ypos-system->win_ypos-reg->hit_yorg+system->yorg;

	/* calculate current row, col */
	new_column=ptrxpos/obj->realwidth;
	if (new_column>=obj->inrow) new_column=obj->inrow-1;
	new_row=ptrypos/itemheight;

	if (new_row==obj->row && new_column==obj->column) return ERR_OK;

	if (new_column*obj->realwidth +xwidth<=ptrxpos &&
		new_row*itemheight+ywidth<=ptrypos)
	{
		if (disp_rows)
		{       /* displayed in rows, old way */
			current=(obj->winyorg+new_row)*obj->inrow+new_column;
			if (current>=obj->count) current=-1;
		}
		else
		{       /* displayed in cols, new way */
			current=obj->winyorg+new_row+new_column*obj->lines_have;
			if (current>=obj->count || new_row+obj->winyorg>=obj->lines_have) current=-1;
		}
	}
	if (obj->current!=-1)
		DrawBorder(obj,&paper); /* undraw old border */

	obj->current=current;
	

	if (current!=-1)
	{
		obj->row	 = new_row;
		obj->column  = new_column;
		DrawBorder(obj,&border);
		if (obj->drag_is_on)
		  return menu_hit_or_done(obj,PW_EVENT_HIT);
	}
	else
	{
		obj->row	= -1;
		obj->column = -1;	/* make sure item is redrawn if ptr goes back in */
	}
	return ERR_OK;
}

/* when pointer enters region -> check to draw a border */
Error Entry(PWObject object)
{
	return pointer_move((Object *)object);
}

Error menu_autosize(Object *obj)
{
	PWSystem *system=obj->event.region->system;
	int lines_want=obj->lines_have;
	int lines_have=(obj->winysiz-ywidth)/itemheight;

	obj->lines_max=max(1,(obj->winysiz+system->Yspaceleft-ywidth)/itemheight);

	if ((lines_have<lines_want && obj->lines_max>lines_have))
	{
		obj->winysiz=min(lines_want,obj->lines_max)*itemheight+ywidth;
		system->status|=SSTATUS_WINCHANGE;
	}

	if (system->Xspaceleft<0 && obj->mstatus&STATUS_MAXWIDTH)
	{
		obj->winxsiz=max(obj->winxsiz+system->Xspaceleft,20);
		system->status|=SSTATUS_WINCHANGE;
		obj->mstatus&=~STATUS_MAXWIDTH;
	}

	return ERR_OK;
}

/* handle pointer move,hit & do */
Error Handle(PWObject object, PWWait event)
{
	Error err=ERR_OK;
	Object *obj=(Object *)object;
	PWSystem *system=obj->event.region->system;

	switch (event)
	{
	case PW_EVENT_MOVE:
		err=pointer_move(obj);
		break;
	case PW_EVENT_HITDRAG:
		obj->drag_is_on=1;
		event=PW_EVENT_HIT;
	case PW_EVENT_HIT:
	case PW_EVENT_DO:
		err=menu_hit_or_done(obj,event);
		break;
	case PW_EVENT_DRAGEND:
		obj->drag_is_on=0;
		break;
	case PW_EVENT_AUTOSIZE:
		menu_autosize(obj);
		break;
	}
	return err;
}

/* define things about object */
void cwidth(void *data)
{
	xwidth=PF2xpix(width); if (xwidth<1) xwidth=1;
	ywidth=PF2ypix(width); if (ywidth<1) ywidth=1;
}

Defines defines[]={
{ "BORDER-WIDTH",               PW_DEFINE_PT,       cwidth, &width,             NULL },
{ "AVAILABLE-INK-COLOUR",       PW_DEFINE_COLOUR,   NULL,   &available.ink,     NULL },
{ "UNAVAILABLE-INK-COLOUR",     PW_DEFINE_COLOUR,   NULL,   &unavailable.ink,   NULL },
{ "SELECTED-INK-COLOUR",        PW_DEFINE_COLOUR,   NULL,   &selected.ink,      NULL },
{ "AVAILABLE-PAPER-COLOUR",     PW_DEFINE_COLOUR,   NULL,   &available.paper,   NULL },
{ "UNAVAILABLE-PAPER-COLOUR",   PW_DEFINE_COLOUR,   NULL,   &unavailable.paper, NULL },
{ "SELECTED-PAPER-COLOUR",      PW_DEFINE_COLOUR,   NULL,   &selected.paper,    NULL },
{ "PAPER-COLOUR",               PW_DEFINE_COLOUR,   NULL,   &paper,             &Kpaper },
{ "INK-COLOUR",                 PW_DEFINE_COLOUR,   NULL,   &ink,               &Kink },
{ "BORDER-COLOUR",              PW_DEFINE_COLOUR,   NULL,   &border,            &Kborder },
{ "FONT",                       PW_DEFINE_FONT,     NULL,   font,               &recalculate },
{ "FONTSIZE",                   PW_DEFINE_PT,       NULL,   &fontsize,          &recalculate },
{ "DISPLAY-ROWS",               PW_DEFINE_BOOLEAN,  NULL,   &disp_rows,         NULL },
{ 0 }
};

Error Define(char *name, char *value)
{
	Error err;
	char *cname="APPLIC-SCROLLBAR-";

	err= PWHandleDefine("MENU-",defines,name,value);
	if (err=ERR_OK) return err;
									
	while (*cname) if (*cname++!=*name++) return ERR_IPAR;

	if (STRSameCD(name,"LEFT"))
	{
		scrollleft=TRUE;
		return ERR_OK;
	}
	if (STRSameCD(name,"RIGHT"))
	{
		scrollleft=FALSE;
		return ERR_OK;
	}
	if (STRSameCD(name,"ABOVE"))
	{
		scrollabove=TRUE;
		return ERR_OK;
	}
	if (STRSameCD(name,"BELOW"))
	{
		scrollabove=FALSE;
		return ERR_OK;
	}
}

/* routine called when object scaled */
Error DoScale(PWObject object, pt xinc, pt yinc)
{
	Error err=ERR_OK;
	Object *obj=(Object *)object;
	PWRegion *reg=obj->event.region;
	pix xpixsiz=reg->hit_xend-reg->hit_xorg;

	obj->winwidth=xpixsiz;			/* set new xsize */
	obj->lines_inwin=(obj->winysiz-ywidth+pt2short(yinc)+itemheight-1)/itemheight;
	return menu_calc(obj);
}

/* possibility to query object */
Error Query(PWObject object, int what, void *_ret)
{
	Error err=ERR_OK;
	Object *obj=(Object *)object;
	PWRegion *reg=obj->event.region;
	char **ret=(char **)_ret;
	int counter;
	char tstat;

	switch (what)
	{
	/* allow querying of the 'canvas' from the scroll bar object*/
	case PW_CANVAS_XORIGIN:
		*((pt *)ret)=obj->winxorg;
		break;
	case PW_CANVAS_YORIGIN:
		*((pt *)ret)=obj->winyorg;
		break;
	case PW_CANVAS_XSIZE:
		/* this code could say : */
		/* *((pt *)ret)=reg->hit_xend-reg->hit_xorg; */
		/* but this doesn't work when called from inside a PWScale event */
		/* As a canvas doesn't have a margin anyway, use (less accurate) */
		*((pt *)ret)=fixmul(reg->tot_xsiz,Xpix2pt);
		break;
	case PW_CANVAS_YSIZE:
		/* this code could say : */
		/* *((pt *)ret)=reg->hit_yend-reg->hit_yorg; */
		/* but this doesn't work when called from inside a PWScale event */
		/* As a canvas doesn't have a margin anyway, use (less accurate) */
		*((pt *)ret)=fixmul(reg->tot_ysiz,Ypix2pt);
		break;

	/* now normal menu query */
	case PW_MENU_STATUS_CURRENT:
		if (obj->qcurrent)
			*(int *)ret=obj->qcurrent->status;
		else
			return ERR_ISYN;
		break;
	case PW_MENU_FIRST:
		obj->qcurrent=obj->base;
	case PW_MENU_NEXT:
		if (obj->qcurrent==NULL) return ERR_ISYN;
		if (obj->qcurrent->next->next)
		{
			obj->qcurrent=obj->qcurrent->next;
			*ret=obj->qcurrent->item;
		}
		else
			*ret=NULL;
		break;
	case PW_MENU_AVAILABLE_FIRST:
		obj->qcurrent=obj->base;
	case PW_MENU_AVAILABLE_NEXT:
		if (obj->qcurrent==NULL) return ERR_ISYN;
		*ret=NULL;
		while (obj->qcurrent->next->next)
		{
			obj->qcurrent=obj->qcurrent->next;
			if (obj->qcurrent->status==PW_STATUS_AVAILABLE)
			{
				*ret=obj->qcurrent->item;
				break;
			}
		}
		break;
	case PW_MENU_SELECTED_FIRST:
		obj->qcurrent=obj->base;
	case PW_MENU_SELECTED_NEXT:
		if (obj->qcurrent==NULL) return ERR_ISYN;
		*ret=NULL;
		while (obj->qcurrent->next->next)
		{
			obj->qcurrent=obj->qcurrent->next;
			if (obj->qcurrent->status==PW_STATUS_SELECTED)
			{
				*ret=obj->qcurrent->item;
				break;
			}
		}
		break;
	case PW_MENU_FIRST_NUMBER:
		obj->qcurrent=obj->base;
	case PW_MENU_NEXT_NUMBER:
		if (obj->qcurrent==NULL) return ERR_ISYN;
		if (obj->qcurrent->next->next)
		{
			Item *item=obj->base->next;
			counter=0;
			obj->qcurrent=obj->qcurrent->next;

			while (item->next && item->item!=obj->qcurrent->item)
			{
				item=item->next;
				counter++;
			}
			if (item->item==obj->qcurrent->item)
				*(int *)ret=counter;
			else
				*(int *)ret=-1; /* oops! */
		}
		else
			*ret=NULL;
		break;

/* THE FOLLOWING RETURN THE NUMBER OF THE SELECTED/AVAILABLE ITEMS */
/* with the following rules: */
/* - numbering starts at 0 (i.e. if the very first menu item is selected */
/*	this returns 0) */
/* - if the end of the elements is reached, this returns -1  */

	case PW_MENU_AVAILABLE_FIRST_NUMBER:
		obj->qcurrent=obj->base;
	case PW_MENU_AVAILABLE_NEXT_NUMBER:
		if (obj->qcurrent==NULL) return ERR_ISYN;
		counter=1;
		if (!obj->qcurrent->next->next)
		{
			*(int *)ret=-1;
			break;
		}
		while (obj->qcurrent->next->next && counter)
		{
			obj->qcurrent=obj->qcurrent->next;
			if (obj->qcurrent->status==PW_STATUS_AVAILABLE)
				counter=0;
		}
		if (obj->qcurrent->status!=PW_STATUS_AVAILABLE)
		{
			*(int *)ret=-1;
			break;
		}
		{
			Item *item=obj->base->next;
			counter=0;
			while (item->next && item->item!=obj->qcurrent->item)
			{
				item=item->next;
				counter++;
			}
		}
		*(int *)ret=counter;
		break;

	case PW_MENU_SELECTED_FIRST_NUMBER:
		obj->qcurrent=obj->base;
	case PW_MENU_SELECTED_NEXT_NUMBER:
		if (obj->qcurrent==NULL) return ERR_ISYN;
		counter=1;
		if (!obj->qcurrent->next->next)
		{
			*(int *)ret=-1;
			break;
		}
		while (obj->qcurrent->next->next && counter)
		{
			obj->qcurrent=obj->qcurrent->next;
			if (obj->qcurrent->status==PW_STATUS_SELECTED)
				counter=0;
		}
		if (obj->qcurrent->status!=PW_STATUS_SELECTED)
		{
			*(int *)ret=-1;
			break;
		}
		{
			Item *item=obj->base->next;
			counter=0;
			while (item->next && item->item!=obj->qcurrent->item)
			{
				item=item->next;
				counter++;
			}
		}
		*(int *)ret=counter;
		break;

	case PW_MENU_SELECTED_NUMBER:
		tstat=PW_STATUS_SELECTED;
		goto nbr_com;
	case PW_MENU_AVAILABLE_NUMBER:
		tstat=PW_STATUS_AVAILABLE;
		goto nbr_com;
	case PW_MENU_UNAVAILABLE_NUMBER:
		tstat=PW_STATUS_UNAVAILABLE;
nbr_com:{
			Item *item=obj->base->next;
			for (counter=0 ; item->next ; item=item->next)
				if (item->status==tstat) counter++;
		}
		*(int *)ret=counter;
		break;

	default:
		err=ERR_IPAR;
		break;
	}
	return err;
}


/* determine size for this object */
Error DetermineSize(PWObject object)
{
	Object *obj=(Object *)object;
	PWRegion *region=obj->event.region;
	region->hit_off_lft=region->hit_off_rgt=
	region->hit_off_top=region->hit_off_bot=0;

	region->tot_xsiz=short2pt(obj->winxsiz);
	region->tot_ysiz=short2pt(obj->winysiz);

	return ERR_OK;
}

PWTypeDef realmenu={
	PW_TYPEDEF_FLAG, PW_TYPE_mENU, PW_TYPE_EVENT_REGION,
	&Create, &Change, &Query, &Remove,
	&Entry, &Exit, &Handle, &Define,
	&Draw, &DetermineSize, &DoScale
};

typedef struct {
	PWBaseObject;					/* common start for all objects */
	PWObject menu;
} dummyObject;

Error dummyCreate(PWObject owner, PWObject *ret, PWEvent event, int *taglist)
{
	event.region->composite=TRUE;

	/* allocate specific object - which should be cleared (automatic) */
	return PWAllocate(event.region->system,sizeof(Object),ret);

	/* no events to react to */
}

Error dummyChange(PWObject object, int *taglist)
{
	dummyObject *obj=(dummyObject *)object;
	if (!(obj->status&OSTATUS_INITIALISED))
	{
		obj->status|=OSTATUS_INITIALISED;		/* init is done */
		return PWDoCreate(object,&obj->menu,PW_TYPE_mENU,taglist);
	}
	else
		return PWDoChange(obj->menu,taglist);
}

Error dummyQuery(PWObject object, int what, void *ret)
{
	dummyObject *obj=(dummyObject *)object;
/*---- change by wl ----*/
	if (what==PW_MENU_BASIC_TRUEMENU)
	{
		*((PWObject *)ret)=obj->menu;
		return ERR_OK;
	}
	else
/* end of change */
	   return PWQuery(obj->menu,what,ret);
}

Error dummyRemove(PWObject object)
{
	return PWRelease(object,object);
}

Error dummyHandle(PWObject object, PWWait event)
{
	dummyObject *obj=(dummyObject *)object;
	return Handle(obj->menu,event);
}

PWTypeDef init={
	(int)&realmenu, PW_TYPE_MENU, PW_TYPE_EVENT_REGION,
	&dummyCreate, &dummyChange, &dummyQuery, &Remove,
	NULL, NULL, &dummyHandle, NULL, NULL, NULL, NULL
};
